/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.core.krendering.extensions;

import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.krendering.ViewSynthesisShared;
import de.cau.cs.kieler.core.math.KVector;
import de.cau.cs.kieler.core.properties.IProperty;
import de.cau.cs.kieler.core.properties.Property;
import de.cau.cs.kieler.core.util.Pair;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import java.util.ArrayList;
import java.util.HashMap;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Provides some helpful extension methods for simplifying the composition of KGraph/KRendering-based view models.<br>
 * <br>
 * In order to employ them beyond KLighD diagram syntheses you best declare a field of type
 * {@link KNodeExtensions} in your class and annotate it with {@link javax.inject.Inject Inject}.<br>
 * <br>
 * Make sure to bind the {@link ViewSynthesisShared} annotation in the employed
 * {@link com.google.inject.Injector Injector} to a {@link com.google.inject.Scope}, e.g. by calling
 * {@code Guice.createInjector(KRenderingExtensionsPlugin.createSingletonScopeBindingModule());} or
 * {@code Guice.createInjector(KRenderingExtensionsPlugin.createNoScopeBindingModule());}.<br>
 * <br>
 * By means of that {@link com.google.inject.Injector Injector} you may get a new instance of your class,
 * or you may inject the above mentioned attribute within instances of your class, e.g. by calling
 * {@code injector.injectMembers(this)} in the constructor of your class.
 * 
 * @author chsch
 * 
 * @containsExtensions
 */
@ViewSynthesisShared
@SuppressWarnings("all")
public class KNodeExtensions {
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  private KNode internalCreateNode(final ArrayList<Object> oc) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(oc);
    final KNode _result;
    synchronized (_createCache_internalCreateNode) {
      if (_createCache_internalCreateNode.containsKey(_cacheKey)) {
        return _createCache_internalCreateNode.get(_cacheKey);
      }
      KNode _createInitializedNode = KimlUtil.createInitializedNode();
      _result = _createInitializedNode;
      _createCache_internalCreateNode.put(_cacheKey, _result);
    }
    _init_internalCreateNode(_result, oc);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KNode> _createCache_internalCreateNode = CollectionLiterals.newHashMap();
  
  private void _init_internalCreateNode(final KNode node, final ArrayList<Object> oc) {
  }
  
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  public KNode getNode(final Object o) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o);
    return this.internalCreateNode(_newArrayList);
  }
  
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  public KNode getNode(final Object o1, final Object o2) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o1, o2);
    return this.internalCreateNode(_newArrayList);
  }
  
  /**
   * A convenience method to create a KNode without relating it to a business object.
   */
  public KNode createNode() {
    return KimlUtil.createInitializedNode();
  }
  
  /**
   * An alias of {@link #getNode} allowing to express in business that the KNode will
   * be created at this place. It is just syntactic sugar.
   */
  public KNode createNode(final Object o) {
    return this.getNode(o);
  }
  
  /**
   * An alias of {@link #getNode} allowing to express in business that the KNode will
   * be created at this place. It is just syntactic sugar.
   */
  public KNode createNode(final Object o1, final Object o2) {
    return this.getNode(o1, o2);
  }
  
  public Pair<Float, Float> getNodeSize(final KNode node) {
    Pair<Float, Float> _pair = new Pair<Float, Float>();
    final Procedure1<Pair<Float, Float>> _function = new Procedure1<Pair<Float, Float>>() {
      public void apply(final Pair<Float, Float> it) {
        final KShapeLayout layout = node.<KShapeLayout>getData(KShapeLayout.class);
        float _height = layout.getHeight();
        it.setFirst(Float.valueOf(_height));
        float _height_1 = layout.getHeight();
        it.setSecond(Float.valueOf(_height_1));
      }
    };
    return ObjectExtensions.<Pair<Float, Float>>operator_doubleArrow(_pair, _function);
  }
  
  public float getHeight(final KNode node) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    return _data.getHeight();
  }
  
  /**
   * Is used in KPortExtensions
   */
  public float getWidth(final KNode node) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    return _data.getWidth();
  }
  
  public KNode setNodeSize(final KNode node, final float width, final float height) {
    final Procedure1<KNode> _function = new Procedure1<KNode>() {
      public void apply(final KNode it) {
        KShapeLayout _data = it.<KShapeLayout>getData(KShapeLayout.class);
        _data.setSize(width, height);
        KNodeExtensions.this.setMinimalNodeSize(it, width, height);
      }
    };
    return ObjectExtensions.<KNode>operator_doubleArrow(node, _function);
  }
  
  public KNode setWidth(final KNode node, final float width) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    final float height = _data.getHeight();
    this.setNodeSize(node, width, height);
    return node;
  }
  
  public KNode setHeight(final KNode node, final float height) {
    KShapeLayout _data = node.<KShapeLayout>getData(KShapeLayout.class);
    final float width = _data.getWidth();
    this.setNodeSize(node, width, height);
    return node;
  }
  
  private final static IProperty<KVector> MINIMAL_NODE_SIZE = new Property<KVector>(
    "de.cau.cs.kieler.klighd.minimalNodeSize", new KVector(10d, 10d));
  
  public KNode setMinimalNodeSize(final KNode node, final float width, final float height) {
    KVector _kVector = new KVector(width, height);
    return this.<KVector>addLayoutParam(node, KNodeExtensions.MINIMAL_NODE_SIZE, _kVector);
  }
  
  public <T extends Object> KNode addLayoutParam(final KNode node, final IProperty<? super T> property, final T value) {
    KShapeLayout _data = null;
    if (node!=null) {
      _data=node.<KShapeLayout>getData(KShapeLayout.class);
    }
    if (_data!=null) {
      _data.<T>setProperty(property, value);
    }
    return node;
  }
}
