/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2014 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.kiml.grana.text;

import com.google.common.base.Objects;
import com.google.common.collect.ContiguousSet;
import com.google.common.collect.DiscreteDomain;
import com.google.common.collect.ImmutableSortedSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.kiml.LayoutMetaDataService;
import de.cau.cs.kieler.kiml.LayoutOptionData;
import de.cau.cs.kieler.kiml.config.CompoundLayoutConfig;
import de.cau.cs.kieler.kiml.config.VolatileLayoutConfig;
import de.cau.cs.kieler.kiml.config.text.LayoutConfigTransformer;
import de.cau.cs.kieler.kiml.grana.AnalysisData;
import de.cau.cs.kieler.kiml.grana.AnalysisService;
import de.cau.cs.kieler.kiml.grana.text.grana.Analysis;
import de.cau.cs.kieler.kiml.grana.text.grana.FloatRange;
import de.cau.cs.kieler.kiml.grana.text.grana.GlobalOutputRef;
import de.cau.cs.kieler.kiml.grana.text.grana.GlobalResourceRef;
import de.cau.cs.kieler.kiml.grana.text.grana.Grana;
import de.cau.cs.kieler.kiml.grana.text.grana.IntRangeRange;
import de.cau.cs.kieler.kiml.grana.text.grana.IntRangeValues;
import de.cau.cs.kieler.kiml.grana.text.grana.Job;
import de.cau.cs.kieler.kiml.grana.text.grana.LocalOutput;
import de.cau.cs.kieler.kiml.grana.text.grana.LocalResource;
import de.cau.cs.kieler.kiml.grana.text.grana.Output;
import de.cau.cs.kieler.kiml.grana.text.grana.OutputReference;
import de.cau.cs.kieler.kiml.grana.text.grana.Range;
import de.cau.cs.kieler.kiml.grana.text.grana.RangeJob;
import de.cau.cs.kieler.kiml.grana.text.grana.RegularJob;
import de.cau.cs.kieler.kiml.grana.text.grana.Resource;
import de.cau.cs.kieler.kiml.grana.text.grana.ResourceReference;
import de.cau.cs.kieler.kiml.grana.ui.batch.Batch;
import de.cau.cs.kieler.kiml.grana.ui.batch.BatchJob;
import de.cau.cs.kieler.kiml.grana.ui.batch.BatchRangeJob;
import de.cau.cs.kieler.kiml.grana.ui.batch.BatchResult;
import de.cau.cs.kieler.kiml.grana.ui.batch.CSVResultSerializer;
import de.cau.cs.kieler.kiml.grana.ui.batch.FileKGraphProvider;
import de.cau.cs.kieler.kiml.grana.ui.batch.IBatchJob;
import de.cau.cs.kieler.kiml.klayoutdata.KIdentifier;
import de.cau.cs.kieler.kiml.service.util.ProgressMonitorAdapter;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.impl.ExtensibleURIConverterImpl;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.StringExtensions;

/**
 * Utility class to convert textually specified grana executions to
 * {@link BatchJob}s and execute them.
 * 
 * Note that there is a slight difference between grana 'Jobs' and grana.text 'Jobs'.
 * To be precise, a grana.text 'Job' corresponds to a grana 'Batch'.
 * 
 * @author uru
 * @kieler.ignore (excluded from review process)
 */
@SuppressWarnings("all")
public final class GranaTextToBatchJob {
  /**
   * @return an iterable with the results of all executed batch jobs.
   *         These can be used to print problems that might have occurred during the execution.
   */
  public static Iterable<BatchResult> execute(final Grana grana, final IProgressMonitor progressMonitor) {
    try {
      final ProgressMonitorAdapter pm = new ProgressMonitorAdapter(progressMonitor);
      IWorkspace _workspace = ResourcesPlugin.getWorkspace();
      final IWorkspaceRoot wsRoot = _workspace.getRoot();
      final LinkedList<BatchResult> results = Lists.<BatchResult>newLinkedList();
      EList<Job> _jobs = grana.getJobs();
      final Function1<Job, Boolean> _function = new Function1<Job, Boolean>() {
        public Boolean apply(final Job it) {
          boolean _or = false;
          boolean _isExecuteAll = grana.isExecuteAll();
          if (_isExecuteAll) {
            _or = true;
          } else {
            EList<Job> _execute = grana.getExecute();
            boolean _contains = _execute.contains(it);
            _or = _contains;
          }
          return Boolean.valueOf(_or);
        }
      };
      Iterable<Job> _filter = IterableExtensions.<Job>filter(_jobs, _function);
      final Function1<Job, Boolean> _function_1 = new Function1<Job, Boolean>() {
        public Boolean apply(final Job it) {
          String _name = it.getName();
          boolean _startsWith = _name.startsWith("_");
          return Boolean.valueOf((!_startsWith));
        }
      };
      Iterable<Job> _filter_1 = IterableExtensions.<Job>filter(_filter, _function_1);
      for (final Job job : _filter_1) {
        {
          EList<Analysis> _analyses = job.getAnalyses();
          final Function1<Analysis, AnalysisData> _function_2 = new Function1<Analysis, AnalysisData>() {
            public AnalysisData apply(final Analysis it) {
              AnalysisService _instance = AnalysisService.getInstance();
              String _name = it.getName();
              return _instance.getAnalysis(_name);
            }
          };
          final List<AnalysisData> analyses = ListExtensions.<Analysis, AnalysisData>map(_analyses, _function_2);
          final Batch batch = new Batch(analyses);
          boolean _matched = false;
          if (!_matched) {
            if (job instanceof RangeJob) {
              _matched=true;
              AnalysisService _instance = AnalysisService.getInstance();
              Analysis _rangeAnalysis = ((RangeJob)job).getRangeAnalysis();
              String _name = _rangeAnalysis.getName();
              final AnalysisData rangeAnalysis = _instance.getAnalysis(_name);
              boolean _equals = Objects.equal(rangeAnalysis, null);
              if (_equals) {
                Analysis _rangeAnalysis_1 = ((RangeJob)job).getRangeAnalysis();
                String _name_1 = _rangeAnalysis_1.getName();
                String _plus = ("Invalid range analysis: " + _name_1);
                throw new IllegalArgumentException(_plus);
              }
              batch.setRangeAnalysis(rangeAnalysis);
              int _rangeAnalysisComponent = ((RangeJob)job).getRangeAnalysisComponent();
              batch.setRangeAnalysisComponent(Integer.valueOf(_rangeAnalysisComponent));
              LayoutMetaDataService _instance_1 = LayoutMetaDataService.getInstance();
              String _rangeOption = ((RangeJob)job).getRangeOption();
              final LayoutOptionData layOpt = _instance_1.getOptionDataBySuffix(_rangeOption);
              boolean _equals_1 = Objects.equal(layOpt, null);
              if (_equals_1) {
                String _rangeOption_1 = ((RangeJob)job).getRangeOption();
                String _plus_1 = ("Invalid layout option: " + _rangeOption_1);
                throw new IllegalArgumentException(_plus_1);
              }
              batch.setRangeOption(layOpt);
              final Range v = ((RangeJob)job).getRangeValues();
              ImmutableSortedSet<? extends Number> _switchResult_1 = null;
              boolean _matched_1 = false;
              if (!_matched_1) {
                if (v instanceof IntRangeRange) {
                  _matched_1=true;
                  int _start = ((IntRangeRange)v).getStart();
                  int _end = ((IntRangeRange)v).getEnd();
                  com.google.common.collect.Range<Integer> _closed = com.google.common.collect.Range.<Integer>closed(Integer.valueOf(_start), Integer.valueOf(_end));
                  DiscreteDomain<Integer> _integers = DiscreteDomain.integers();
                  _switchResult_1 = ContiguousSet.<Integer>create(_closed, _integers);
                }
              }
              if (!_matched_1) {
                if (v instanceof IntRangeValues) {
                  _matched_1=true;
                  EList<Integer> _values = ((IntRangeValues)v).getValues();
                  _switchResult_1 = ContiguousSet.<Integer>copyOf(_values);
                }
              }
              if (!_matched_1) {
                if (v instanceof FloatRange) {
                  _matched_1=true;
                  EList<Float> _values = ((FloatRange)v).getValues();
                  _switchResult_1 = ContiguousSet.<Float>copyOf(_values);
                }
              }
              final ImmutableSortedSet<? extends Number> vals = _switchResult_1;
              batch.setRangeValues(vals);
            }
          }
          EList<Resource> _resources = job.getResources();
          final Function1<Resource, Iterable<LocalResource>> _function_3 = new Function1<Resource, Iterable<LocalResource>>() {
            public Iterable<LocalResource> apply(final Resource r) {
              Iterable<LocalResource> _switchResult = null;
              boolean _matched = false;
              if (!_matched) {
                if (r instanceof ResourceReference) {
                  _matched=true;
                  EList<GlobalResourceRef> _resourceRefs = ((ResourceReference) r).getResourceRefs();
                  final Function1<GlobalResourceRef, EList<LocalResource>> _function = new Function1<GlobalResourceRef, EList<LocalResource>>() {
                    public EList<LocalResource> apply(final GlobalResourceRef it) {
                      return it.getResources();
                    }
                  };
                  List<EList<LocalResource>> _map = ListExtensions.<GlobalResourceRef, EList<LocalResource>>map(_resourceRefs, _function);
                  _switchResult = Iterables.<LocalResource>concat(_map);
                }
              }
              if (!_matched) {
                _switchResult = Collections.<LocalResource>unmodifiableList(CollectionLiterals.<LocalResource>newArrayList(((LocalResource) r)));
              }
              return _switchResult;
            }
          };
          List<Iterable<LocalResource>> _map = ListExtensions.<Resource, Iterable<LocalResource>>map(_resources, _function_3);
          final Iterable<LocalResource> resources = Iterables.<LocalResource>concat(_map);
          for (final LocalResource resource : resources) {
            {
              Pattern _xifexpression = null;
              String _filter_2 = resource.getFilter();
              boolean _isNullOrEmpty = StringExtensions.isNullOrEmpty(_filter_2);
              boolean _not = (!_isNullOrEmpty);
              if (_not) {
                String _filter_3 = resource.getFilter();
                _xifexpression = Pattern.compile(_filter_3);
              } else {
                _xifexpression = null;
              }
              final Pattern filter = _xifexpression;
              try {
                IProject[] _projects = wsRoot.getProjects();
                final Function1<IProject, Boolean> _function_4 = new Function1<IProject, Boolean>() {
                  public Boolean apply(final IProject p) {
                    String _path = resource.getPath();
                    String _name = p.getName();
                    return Boolean.valueOf(_path.contains(_name));
                  }
                };
                final IProject p = IterableExtensions.<IProject>findFirst(((Iterable<IProject>)Conversions.doWrapArray(_projects)), _function_4);
                boolean _equals = Objects.equal(p, null);
                if (_equals) {
                  String _path = resource.getPath();
                  String _plus = ("Could not find resource " + _path);
                  throw new IllegalArgumentException(_plus);
                }
                String _path_1 = resource.getPath();
                String _name = p.getName();
                String _replace = _path_1.replace(_name, "");
                final IResource wsloc = p.findMember(_replace);
                IResource[] _members = ((IContainer) wsloc).members();
                for (final IResource file : _members) {
                  boolean _or = false;
                  boolean _equals_1 = Objects.equal(filter, null);
                  if (_equals_1) {
                    _or = true;
                  } else {
                    String _name_1 = file.getName();
                    Matcher _matcher = filter.matcher(_name_1);
                    boolean _matches = _matcher.matches();
                    _or = _matches;
                  }
                  if (_or) {
                    String _path_2 = resource.getPath();
                    String _plus_1 = (_path_2 + "/");
                    String _name_2 = file.getName();
                    String _plus_2 = (_plus_1 + _name_2);
                    Path _path_3 = new Path(_plus_2);
                    GranaTextToBatchJob.addBatchJob(batch, job, _path_3);
                  }
                }
              } catch (final Throwable _t) {
                if (_t instanceof Exception) {
                  final Exception e = (Exception)_t;
                  String _path_4 = resource.getPath();
                  final URI fileURI = URI.createURI(_path_4, true);
                  String _fileString = fileURI.toFileString();
                  final File dir = new File(_fileString);
                  boolean _or_1 = false;
                  boolean _equals_2 = Objects.equal(dir, null);
                  if (_equals_2) {
                    _or_1 = true;
                  } else {
                    boolean _exists = dir.exists();
                    boolean _not_1 = (!_exists);
                    _or_1 = _not_1;
                  }
                  if (_or_1) {
                    String _path_5 = resource.getPath();
                    String _plus_3 = ("Could not find resource location: \'" + _path_5);
                    String _plus_4 = (_plus_3 + "\'");
                    throw new IllegalArgumentException(_plus_4);
                  }
                  File[] _listFiles = dir.listFiles();
                  for (final File file_1 : _listFiles) {
                    boolean _or_2 = false;
                    boolean _equals_3 = Objects.equal(filter, null);
                    if (_equals_3) {
                      _or_2 = true;
                    } else {
                      String _name_3 = file_1.getName();
                      Matcher _matcher_1 = filter.matcher(_name_3);
                      boolean _matches_1 = _matcher_1.matches();
                      _or_2 = _matches_1;
                    }
                    if (_or_2) {
                      String _absolutePath = file_1.getAbsolutePath();
                      Path _path_6 = new Path(_absolutePath);
                      GranaTextToBatchJob.addBatchJob(batch, job, _path_6);
                    }
                  }
                } else {
                  throw Exceptions.sneakyThrow(_t);
                }
              }
            }
          }
          String _switchResult_1 = null;
          Output _output = job.getOutput();
          boolean _matched_1 = false;
          if (!_matched_1) {
            if (_output instanceof OutputReference) {
              _matched_1=true;
              Output _output_1 = job.getOutput();
              GlobalOutputRef _outputRef = ((OutputReference) _output_1).getOutputRef();
              LocalOutput _output_2 = _outputRef.getOutput();
              _switchResult_1 = _output_2.getPath();
            }
          }
          if (!_matched_1) {
            Output _output_1 = job.getOutput();
            _switchResult_1 = ((LocalOutput) _output_1).getPath();
          }
          final String outputPath = _switchResult_1;
          try {
            OutputStream os = null;
            final URI fileURI = URI.createURI(outputPath, true);
            final ExtensibleURIConverterImpl uriConv = new ExtensibleURIConverterImpl();
            OutputStream _createOutputStream = uriConv.createOutputStream(fileURI);
            os = _createOutputStream;
            os.write(0);
            os.close();
          } catch (final Throwable _t) {
            if (_t instanceof IOException) {
              final IOException e = (IOException)_t;
              throw new IllegalArgumentException(
                (("Could not store results to file \'" + outputPath) + "\'"));
            } else {
              throw Exceptions.sneakyThrow(_t);
            }
          }
          IKielerProgressMonitor _subTask = pm.subTask(1);
          final BatchResult result = batch.execute(_subTask);
          boolean _isCanceled = pm.isCanceled();
          if (_isCanceled) {
            return Collections.<BatchResult>unmodifiableList(CollectionLiterals.<BatchResult>newArrayList());
          }
          try {
            OutputStream os_1 = null;
            final URI fileURI_1 = URI.createURI(outputPath, true);
            final ExtensibleURIConverterImpl uriConv_1 = new ExtensibleURIConverterImpl();
            OutputStream _createOutputStream_1 = uriConv_1.createOutputStream(fileURI_1);
            os_1 = _createOutputStream_1;
            final CSVResultSerializer serializer = new CSVResultSerializer();
            IKielerProgressMonitor _subTask_1 = pm.subTask(1);
            serializer.serialize(os_1, result, _subTask_1);
            os_1.close();
          } catch (final Throwable _t_1) {
            if (_t_1 instanceof IOException) {
              final IOException e_1 = (IOException)_t_1;
              throw new IllegalArgumentException(
                (("Could not store results to file \'" + outputPath) + "\'"));
            } else {
              throw Exceptions.sneakyThrow(_t_1);
            }
          }
          results.add(result);
        }
      }
      return results;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  /**
   * @param batch
   *          grana's current batch instance
   * @param job
   *          grana.text's Job corresponding to the batch
   * @param path
   *          the path of a test case (corresponding to a grana Job)
   */
  private static void addBatchJob(final Batch batch, final Job job, final IPath path) {
    final FileKGraphProvider provider = new FileKGraphProvider();
    boolean _matched = false;
    if (!_matched) {
      if (job instanceof RegularJob) {
        _matched=true;
        boolean _isLayoutBeforeAnalysis = ((RegularJob)job).isLayoutBeforeAnalysis();
        provider.setLayoutBeforeAnalysis(_isLayoutBeforeAnalysis);
        boolean _isMeasureExecutionTime = ((RegularJob)job).isMeasureExecutionTime();
        provider.setExecutionTimeAnalysis(_isMeasureExecutionTime);
      }
    }
    if (!_matched) {
      if (job instanceof RangeJob) {
        _matched=true;
        provider.setLayoutBeforeAnalysis(true);
        provider.setExecutionTimeAnalysis(false);
      }
    }
    final CompoundLayoutConfig clc = new CompoundLayoutConfig();
    EList<KIdentifier> _layoutOptions = job.getLayoutOptions();
    for (final KIdentifier cfg : _layoutOptions) {
      VolatileLayoutConfig _transformConfig = LayoutConfigTransformer.transformConfig(cfg);
      clc.add(_transformConfig);
    }
    provider.setLayoutConfigurator(clc);
    IBatchJob<IPath> _switchResult_1 = null;
    boolean _matched_1 = false;
    if (!_matched_1) {
      if (job instanceof RegularJob) {
        _matched_1=true;
        _switchResult_1 = new BatchJob<IPath>(path, provider);
      }
    }
    if (!_matched_1) {
      if (job instanceof RangeJob) {
        _matched_1=true;
        _switchResult_1 = new BatchRangeJob<IPath>(batch, path, provider);
      }
    }
    final IBatchJob<IPath> batchJob = _switchResult_1;
    batch.appendJob(batchJob);
  }
}
