/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.klighd.examples.circuit;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KLabel;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.krendering.KArc;
import de.cau.cs.kieler.core.krendering.KAreaPlacementData;
import de.cau.cs.kieler.core.krendering.KColor;
import de.cau.cs.kieler.core.krendering.KEllipse;
import de.cau.cs.kieler.core.krendering.KLineCap;
import de.cau.cs.kieler.core.krendering.KPolygon;
import de.cau.cs.kieler.core.krendering.KPolyline;
import de.cau.cs.kieler.core.krendering.KRectangle;
import de.cau.cs.kieler.core.krendering.KRendering;
import de.cau.cs.kieler.core.krendering.KRenderingFactory;
import de.cau.cs.kieler.core.krendering.KRoundedBendsPolyline;
import de.cau.cs.kieler.core.krendering.KRoundedRectangle;
import de.cau.cs.kieler.core.krendering.LineCap;
import de.cau.cs.kieler.core.krendering.LineJoin;
import de.cau.cs.kieler.core.krendering.extensions.KColorExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KContainerRenderingExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KEdgeExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KLabelExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KNodeExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KPolylineExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KPortExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.options.PortSide;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.klighd.examples.circuit.Circuit;
import de.cau.cs.kieler.klighd.examples.circuit.Connector;
import de.cau.cs.kieler.klighd.examples.circuit.Wire;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import javax.inject.Inject;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtext.xbase.lib.CollectionExtensions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * An exemplary diagram synthesis illustrating the generation of diagrams containing nodes with ports.
 * 
 * @author chsch
 */
@SuppressWarnings("all")
public class CircuitDiagramSynthesis extends AbstractDiagramSynthesis<Circuit> {
  @Inject
  @Extension
  private KNodeExtensions _kNodeExtensions;
  
  @Inject
  @Extension
  private KEdgeExtensions _kEdgeExtensions;
  
  @Inject
  @Extension
  private KPortExtensions _kPortExtensions;
  
  @Inject
  @Extension
  private KLabelExtensions _kLabelExtensions;
  
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  @Inject
  @Extension
  private KContainerRenderingExtensions _kContainerRenderingExtensions;
  
  @Inject
  @Extension
  private KPolylineExtensions _kPolylineExtensions;
  
  @Inject
  @Extension
  private KColorExtensions _kColorExtensions;
  
  @Extension
  private KRenderingFactory _kRenderingFactory = KRenderingFactory.eINSTANCE;
  
  /**
   * {@inheritDoc}<br>
   * <br>
   * This method is called by KLighD for synthesizing a KGraph/KRendering-based view model
   * describing a representation of the given {@link Circuit} <code>circuit</code>.
   */
  public KNode transform(final Circuit circuit) {
    KNode _createNode = this._kNodeExtensions.createNode();
    final KNode diagramRoot = this.<KNode>associateWith(_createNode, circuit);
    this.createCircuitNode(circuit, diagramRoot);
    return diagramRoot;
  }
  
  /**
   * This method translates instances of {@link Circuit} recursively into {@link KNode KNodes} and
   * adds it to the given <code>parent</code> node.
   * 
   * @param circuit
   *            the circuit to be translated
   * @param parent
   *            the parent {@link KNode} the new one is to be attached.
   */
  public void createCircuitNode(final Circuit circuit, final KNode parent) {
    KNode _createNode = this._kNodeExtensions.createNode(circuit);
    final KNode circuitNode = this.<KNode>associateWith(_createNode, circuit);
    EList<KNode> _children = parent.getChildren();
    _children.add(circuitNode);
    this.<KNode, EdgeRouting>setLayoutOption(circuitNode, LayoutOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);
    EList<Circuit> _innerCircuits = circuit.getInnerCircuits();
    final boolean atomicCircuit = _innerCircuits.isEmpty();
    EList<Connector> _connectors = circuit.getConnectors();
    final Procedure1<Connector> _function = new Procedure1<Connector>() {
      public void apply(final Connector connector) {
        KPort _createPort = CircuitDiagramSynthesis.this._kPortExtensions.createPort(connector);
        KPort _associateWith = CircuitDiagramSynthesis.this.<KPort>associateWith(_createPort, connector);
        final Procedure1<KPort> _function = new Procedure1<KPort>() {
          public void apply(final KPort it) {
            EList<KPort> _ports = circuitNode.getPorts();
            _ports.add(it);
            CircuitDiagramSynthesis.this._kPortExtensions.setPortSize(it, 5, 2);
            float _xifexpression = (float) 0;
            if (atomicCircuit) {
              _xifexpression = 0f;
            } else {
              _xifexpression = (-3f);
            }
            CircuitDiagramSynthesis.this.<KPort, Float>setLayoutOption(it, LayoutOptions.OFFSET, Float.valueOf(_xifexpression));
            KRectangle _addRectangle = CircuitDiagramSynthesis.this._kRenderingExtensions.addRectangle(it);
            KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("black");
            KRectangle _setBackground = CircuitDiagramSynthesis.this._kRenderingExtensions.<KRectangle>setBackground(_addRectangle, _color);
            CircuitDiagramSynthesis.this._kRenderingExtensions.setLineJoin(_setBackground, LineJoin.JOIN_ROUND);
            String _name = connector.getName();
            KLabel _addInsidePortLabel = CircuitDiagramSynthesis.this._kLabelExtensions.addInsidePortLabel(it, _name, 8, KlighdConstants.DEFAULT_FONT_NAME);
            CircuitDiagramSynthesis.this.<KLabel>associateWith(_addInsidePortLabel, connector);
          }
        };
        ObjectExtensions.<KPort>operator_doubleArrow(_associateWith, _function);
      }
    };
    IterableExtensions.<Connector>forEach(_connectors, _function);
    String _type = circuit.getType();
    boolean _matched = false;
    if (!_matched) {
      if (Objects.equal(_type, "NOT")) {
        _matched=true;
        this.createNotGateRendering(circuitNode);
      }
    }
    if (!_matched) {
      if (Objects.equal(_type, "AND")) {
        _matched=true;
        this.createAndGateRendering(circuitNode);
      }
    }
    if (!_matched) {
      if (Objects.equal(_type, "OR")) {
        _matched=true;
        this.createOrGateRendering(circuitNode);
      }
    }
    if (!_matched) {
      if (atomicCircuit) {
        _matched=true;
        this.createBasicGateRendering(circuitNode);
      }
    }
    if (!_matched) {
      KNode _setNodeSize = this._kNodeExtensions.setNodeSize(circuitNode, 40, 40);
      KNode _setExpansionAwareLayoutOption = this.<PortConstraints>setExpansionAwareLayoutOption(_setNodeSize, 
        LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_ORDER, PortConstraints.FREE);
      KRoundedRectangle _addRoundedRectangle = this._kRenderingExtensions.addRoundedRectangle(_setExpansionAwareLayoutOption, 3, 5);
      final Procedure1<KRoundedRectangle> _function_1 = new Procedure1<KRoundedRectangle>() {
        public void apply(final KRoundedRectangle it) {
          KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("black");
          CircuitDiagramSynthesis.this._kRenderingExtensions.setShadow(it, _color);
          KColor _color_1 = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
          CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color_1);
          CircuitDiagramSynthesis.this._kRenderingExtensions.<KRoundedRectangle>addDoubleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
        }
      };
      ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_addRoundedRectangle, _function_1);
    }
    EList<Circuit> _innerCircuits_1 = circuit.getInnerCircuits();
    final Procedure1<Circuit> _function_2 = new Procedure1<Circuit>() {
      public void apply(final Circuit it) {
        CircuitDiagramSynthesis.this.createCircuitNode(it, circuitNode);
      }
    };
    IterableExtensions.<Circuit>forEach(_innerCircuits_1, _function_2);
    EList<Wire> _innerWires = circuit.getInnerWires();
    final Procedure1<Wire> _function_3 = new Procedure1<Wire>() {
      public void apply(final Wire wire) {
        EList<Connector> _connectedTo = wire.getConnectedTo();
        Iterable<Connector> _tail = IterableExtensions.<Connector>tail(_connectedTo);
        final Procedure1<Connector> _function = new Procedure1<Connector>() {
          public void apply(final Connector connector) {
            KEdge _createEdge = CircuitDiagramSynthesis.this._kEdgeExtensions.createEdge();
            KEdge _associateWith = CircuitDiagramSynthesis.this.<KEdge>associateWith(_createEdge, wire);
            final Procedure1<KEdge> _function = new Procedure1<KEdge>() {
              public void apply(final KEdge it) {
                EList<Connector> _connectedTo = wire.getConnectedTo();
                Connector _head = IterableExtensions.<Connector>head(_connectedTo);
                Circuit _parent = null;
                if (_head!=null) {
                  _parent=_head.getParent();
                }
                KNode _node = null;
                if (_parent!=null) {
                  _node=CircuitDiagramSynthesis.this._kNodeExtensions.getNode(_parent);
                }
                it.setSource(_node);
                EList<Connector> _connectedTo_1 = wire.getConnectedTo();
                Connector _head_1 = IterableExtensions.<Connector>head(_connectedTo_1);
                KPort _port = null;
                if (_head_1!=null) {
                  _port=CircuitDiagramSynthesis.this._kPortExtensions.getPort(_head_1);
                }
                it.setSourcePort(_port);
                Circuit _parent_1 = connector.getParent();
                KNode _node_1 = null;
                if (_parent_1!=null) {
                  _node_1=CircuitDiagramSynthesis.this._kNodeExtensions.getNode(_parent_1);
                }
                it.setTarget(_node_1);
                KPort _port_1 = CircuitDiagramSynthesis.this._kPortExtensions.getPort(connector);
                it.setTargetPort(_port_1);
                KRoundedBendsPolyline _addRoundedBendsPolyline = CircuitDiagramSynthesis.this._kEdgeExtensions.addRoundedBendsPolyline(it, 3);
                CircuitDiagramSynthesis.this._kPolylineExtensions.addJunctionPointDecorator(_addRoundedBendsPolyline);
              }
            };
            ObjectExtensions.<KEdge>operator_doubleArrow(_associateWith, _function);
          }
        };
        IterableExtensions.<Connector>forEach(_tail, _function);
      }
    };
    IterableExtensions.<Wire>forEach(_innerWires, _function_3);
  }
  
  public KRendering createBasicGateRendering(final KNode node) {
    KRectangle _xblockexpression = null;
    {
      this._kNodeExtensions.setNodeSize(node, 40, 40);
      KRectangle _addRectangle = this._kRenderingExtensions.addRectangle(node);
      final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
        public void apply(final KRectangle it) {
          CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
          CircuitDiagramSynthesis.this._kRenderingExtensions.setLineCap(it, LineCap.CAP_ROUND);
          CircuitDiagramSynthesis.this._kRenderingExtensions.setLineJoin(it, LineJoin.JOIN_ROUND);
          KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("white");
          CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, _color);
          KColor _color_1 = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
          CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color_1);
        }
      };
      _xblockexpression = ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
    }
    return _xblockexpression;
  }
  
  private final KColor customLightBlue = this._kRenderingFactory.createKColor().setColor(226, 237, 255);
  
  public KRendering createNotGateRendering(final KNode node) {
    KRectangle _xblockexpression = null;
    {
      this._kNodeExtensions.setNodeSize(node, 30, 30);
      this.<KNode, PortConstraints>setLayoutOption(node, LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_SIDE);
      EList<KPort> _ports = node.getPorts();
      KPort _head = IterableExtensions.<KPort>head(_ports);
      this.<KPort, PortSide>setLayoutOption(_head, LayoutOptions.PORT_SIDE, PortSide.WEST);
      EList<KPort> _ports_1 = node.getPorts();
      KPort _last = IterableExtensions.<KPort>last(_ports_1);
      final Procedure1<KPort> _function = new Procedure1<KPort>() {
        public void apply(final KPort it) {
          CircuitDiagramSynthesis.this.<KPort, PortSide>setLayoutOption(it, LayoutOptions.PORT_SIDE, PortSide.EAST);
          CircuitDiagramSynthesis.this.<KPort, Float>setLayoutOption(it, LayoutOptions.OFFSET, Float.valueOf(0f));
          CircuitDiagramSynthesis.this._kPortExtensions.setPortSize(it, 5, 5);
          EList<KGraphData> _data = it.getData();
          EList<KGraphData> _data_1 = it.getData();
          Iterable<KRendering> _filter = Iterables.<KRendering>filter(_data_1, KRendering.class);
          CollectionExtensions.<KGraphData>removeAll(_data, _filter);
          KEllipse _addEllipse = CircuitDiagramSynthesis.this._kRenderingExtensions.addEllipse(it);
          KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("white");
          KEllipse _setBackground = CircuitDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackground(_addEllipse, _color);
          CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(_setBackground, 1.5f);
        }
      };
      ObjectExtensions.<KPort>operator_doubleArrow(_last, _function);
      KRectangle _addRectangle = this._kRenderingExtensions.addRectangle(node);
      final Procedure1<KRectangle> _function_1 = new Procedure1<KRectangle>() {
        public void apply(final KRectangle it) {
          CircuitDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
          KPolygon _addPolygon = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addPolygon(it);
          final Procedure1<KPolygon> _function = new Procedure1<KPolygon>() {
            public void apply(final KPolygon it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineCap(it, LineCap.CAP_ROUND);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineJoin(it, LineJoin.JOIN_ROUND);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, CircuitDiagramSynthesis.this.customLightBlue);
              KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
              CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 1, 0);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0.5f);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 1, 0);
            }
          };
          ObjectExtensions.<KPolygon>operator_doubleArrow(_addPolygon, _function);
        }
      };
      _xblockexpression = ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function_1);
    }
    return _xblockexpression;
  }
  
  public KRendering createAndGateRendering(final KNode node) {
    KRectangle _xblockexpression = null;
    {
      this._kNodeExtensions.setNodeSize(node, 40, 30);
      this.<KNode, PortConstraints>setLayoutOption(node, LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_SIDE);
      EList<KPort> _ports = node.getPorts();
      final Procedure1<KPort> _function = new Procedure1<KPort>() {
        public void apply(final KPort it) {
          CircuitDiagramSynthesis.this.<KPort, PortSide>setLayoutOption(it, LayoutOptions.PORT_SIDE, PortSide.WEST);
        }
      };
      IterableExtensions.<KPort>forEach(_ports, _function);
      EList<KPort> _ports_1 = node.getPorts();
      KPort _last = IterableExtensions.<KPort>last(_ports_1);
      KPort _setLayoutOption = this.<KPort, PortSide>setLayoutOption(_last, LayoutOptions.PORT_SIDE, PortSide.EAST);
      this.<KPort, Float>setLayoutOption(_setLayoutOption, LayoutOptions.OFFSET, Float.valueOf((-1f)));
      KRectangle _addRectangle = this._kRenderingExtensions.addRectangle(node);
      final Procedure1<KRectangle> _function_1 = new Procedure1<KRectangle>() {
        public void apply(final KRectangle it) {
          CircuitDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
          KRectangle _addRectangle = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
          final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
            public void apply(final KRectangle it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 0);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, CircuitDiagramSynthesis.this.customLightBlue);
              KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
              CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color);
              KAreaPlacementData _setAreaPlacementData = CircuitDiagramSynthesis.this._kRenderingExtensions.setAreaPlacementData(it);
              KAreaPlacementData _from = CircuitDiagramSynthesis.this._kRenderingExtensions.from(_setAreaPlacementData, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0);
              CircuitDiagramSynthesis.this._kRenderingExtensions.to(_from, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 14f, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0);
            }
          };
          ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
          KPolyline _addPolyline = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addPolyline(it);
          final Procedure1<KPolyline> _function_1 = new Procedure1<KPolyline>() {
            public void apply(final KPolyline it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineCap(it, LineCap.CAP_ROUND);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineJoin(it, LineJoin.JOIN_ROUND);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 14.5f, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 1, 0);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 1, 0);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 1, 0);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 14.5f, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 1, 0);
            }
          };
          ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function_1);
          KArc _addArc = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addArc(it);
          final Procedure1<KArc> _function_2 = new Procedure1<KArc>() {
            public void apply(final KArc it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineCap(it, LineCap.CAP_ROUND);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, CircuitDiagramSynthesis.this.customLightBlue);
              KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
              CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color);
              it.setArcAngle(180);
              it.setStartAngle((-90));
              KAreaPlacementData _setAreaPlacementData = CircuitDiagramSynthesis.this._kRenderingExtensions.setAreaPlacementData(it);
              CircuitDiagramSynthesis.this._kRenderingExtensions.from(_setAreaPlacementData, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 10, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0);
            }
          };
          ObjectExtensions.<KArc>operator_doubleArrow(_addArc, _function_2);
        }
      };
      _xblockexpression = ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function_1);
    }
    return _xblockexpression;
  }
  
  public KRendering createOrGateRendering(final KNode node) {
    KRectangle _xblockexpression = null;
    {
      this._kNodeExtensions.setNodeSize(node, 40, 30);
      this.<KNode, PortConstraints>setLayoutOption(node, LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_SIDE);
      EList<KPort> _ports = node.getPorts();
      final Procedure1<KPort> _function = new Procedure1<KPort>() {
        public void apply(final KPort it) {
          KPort _setLayoutOption = CircuitDiagramSynthesis.this.<KPort, PortSide>setLayoutOption(it, LayoutOptions.PORT_SIDE, PortSide.WEST);
          CircuitDiagramSynthesis.this.<KPort, Float>setLayoutOption(_setLayoutOption, LayoutOptions.OFFSET, Float.valueOf((-8f)));
        }
      };
      IterableExtensions.<KPort>forEach(_ports, _function);
      EList<KPort> _ports_1 = node.getPorts();
      KPort _last = IterableExtensions.<KPort>last(_ports_1);
      KPort _setLayoutOption = this.<KPort, PortSide>setLayoutOption(_last, LayoutOptions.PORT_SIDE, PortSide.EAST);
      this.<KPort, Float>setLayoutOption(_setLayoutOption, LayoutOptions.OFFSET, Float.valueOf((-1f)));
      KRectangle _addRectangle = this._kRenderingExtensions.addRectangle(node);
      final Procedure1<KRectangle> _function_1 = new Procedure1<KRectangle>() {
        public void apply(final KRectangle it) {
          CircuitDiagramSynthesis.this._kRenderingExtensions.setInvisible(it, true);
          CircuitDiagramSynthesis.this._kRenderingExtensions.setLineCap(it, LineCap.CAP_ROUND);
          KLineCap _lineCap = CircuitDiagramSynthesis.this._kRenderingExtensions.getLineCap(it);
          _lineCap.setPropagateToChildren(true);
          KRectangle _addRectangle = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addRectangle(it);
          final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
            public void apply(final KRectangle it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 0);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, CircuitDiagramSynthesis.this.customLightBlue);
              KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
              CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color);
              KAreaPlacementData _setAreaPlacementData = CircuitDiagramSynthesis.this._kRenderingExtensions.setAreaPlacementData(it);
              KAreaPlacementData _from = CircuitDiagramSynthesis.this._kRenderingExtensions.from(_setAreaPlacementData, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0.1f, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0);
              CircuitDiagramSynthesis.this._kRenderingExtensions.to(_from, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 14f, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0);
            }
          };
          ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function);
          KPolyline _addPolyline = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addPolyline(it);
          final Procedure1<KPolyline> _function_1 = new Procedure1<KPolyline>() {
            public void apply(final KPolyline it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 1, 0);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 15, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 1, 0);
            }
          };
          ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function_1);
          KPolyline _addPolyline_1 = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addPolyline(it);
          final Procedure1<KPolyline> _function_2 = new Procedure1<KPolyline>() {
            public void apply(final KPolyline it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 0, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 1, 0);
              CircuitDiagramSynthesis.this._kPolylineExtensions.addKPosition(it, CircuitDiagramSynthesis.this._kRenderingExtensions.RIGHT, 15, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 1, 0);
            }
          };
          ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline_1, _function_2);
          KArc _addArc = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addArc(it);
          final Procedure1<KArc> _function_3 = new Procedure1<KArc>() {
            public void apply(final KArc it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              it.setArcAngle(180);
              it.setStartAngle((-90));
              KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("white");
              CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, _color);
              KAreaPlacementData _setAreaPlacementData = CircuitDiagramSynthesis.this._kRenderingExtensions.setAreaPlacementData(it);
              KAreaPlacementData _from = CircuitDiagramSynthesis.this._kRenderingExtensions.from(_setAreaPlacementData, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, (-10.1f), 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0);
              CircuitDiagramSynthesis.this._kRenderingExtensions.to(_from, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 10, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.BOTTOM, 0, 0);
            }
          };
          ObjectExtensions.<KArc>operator_doubleArrow(_addArc, _function_3);
          KArc _addArc_1 = CircuitDiagramSynthesis.this._kContainerRenderingExtensions.addArc(it);
          final Procedure1<KArc> _function_4 = new Procedure1<KArc>() {
            public void apply(final KArc it) {
              CircuitDiagramSynthesis.this._kRenderingExtensions.setLineWidth(it, 2);
              CircuitDiagramSynthesis.this._kRenderingExtensions.setBackground(it, CircuitDiagramSynthesis.this.customLightBlue);
              KColor _color = CircuitDiagramSynthesis.this._kColorExtensions.getColor("gray");
              CircuitDiagramSynthesis.this._kRenderingExtensions.setSelectionBackground(it, _color);
              it.setArcAngle(180);
              it.setStartAngle((-90));
              KAreaPlacementData _setAreaPlacementData = CircuitDiagramSynthesis.this._kRenderingExtensions.setAreaPlacementData(it);
              CircuitDiagramSynthesis.this._kRenderingExtensions.from(_setAreaPlacementData, CircuitDiagramSynthesis.this._kRenderingExtensions.LEFT, 10, 0, CircuitDiagramSynthesis.this._kRenderingExtensions.TOP, 0, 0);
            }
          };
          ObjectExtensions.<KArc>operator_doubleArrow(_addArc_1, _function_4);
        }
      };
      _xblockexpression = ObjectExtensions.<KRectangle>operator_doubleArrow(_addRectangle, _function_1);
    }
    return _xblockexpression;
  }
}
