/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.krendering.KRendering;
import de.cau.cs.kieler.core.util.Pair;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klighd.microlayout.Bounds;
import de.cau.cs.kieler.klighd.microlayout.PlacementUtil;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.KRenderingFigureProvider;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.FeatureMap;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.ptolemy.moml.ConfigureType;
import org.ptolemy.moml.PropertyType;

/**
 * Extracts comments from the model and turns them into special comment nodes. Also tries to find the
 * entities the comments probably refer to and attaches them to these entities.
 * 
 * <p>In Ptolemy, certain annotations in a model are like comments in source code.
 * There are two ways how they can be represented in MOML:</p>
 * 
 * <ol>
 *   <li>Using a property of type "ptolemy.vergil.kernel.attributes.TextAttribute"
 *       with another property named "text", as follows:
 *     <pre>
 *        &lt;property
 *              name="Annotation2"
 *              class="ptolemy.vergil.kernel.attributes.TextAttribute"&gt;
 * 
 *              &lt;property
 *                      name="text"
 *                      class="ptolemy.kernel.util.StringAttribute"
 *                      value="This is my annotation's text."&gt;
 *              &lt;/property&gt;
 *              &lt;property
 *                      name="_location"
 *                      class="ptolemy.kernel.util.Location"
 *                      value="[140.0, 440.0]"&gt;
 *              &lt;/property&gt;
 *        &lt;/property&gt;
 *     </pre>
 *   </li>
 * 
 *   <li>Using a property of type "ptolemy.kernel.util.Attribute" with an SVG
 *       as its "_iconDescription" property, as follows:
 *     <pre>
 *        &lt;property
 *              name="annotation3"
 *              class="ptolemy.kernel.util.Attribute"&gt;
 * 
 *              &lt;property
 *                      name="_iconDescription"
 *                      class="ptolemy.kernel.util.SingletonConfigurableAttribute"&gt;
 * 
 *                      &lt;configure&gt;&lt;svg&gt;
 *                              &lt;text x="20" y="20" style="..."&gt;
 *                                      This is my annotation's text.
 *                              &lt;/text&gt;
 *                      &lt;/svg&gt;&lt;/configure&gt;
 *              &lt;/property&gt;
 *              &lt;property
 *                      name="_location"
 *                      class="ptolemy.kernel.util.Location"
 *                      value="[325.0, 10.0]"&gt;
 *              &lt;/property&gt;
 *        &lt;/property&gt;
 *     </pre>
 *   </li>
 * </ol>
 * 
 * <p>while the first version is straightforward, the latter version causes a whole
 * lot of problems. The {@code configure} element is a mixed
 * element, which means that it can contain anything, not just XML. However, it
 * does contain XML (usually an {@code svg} element and its children), which
 * disturbs the quiet peace of the parser. (which, in turn, disturbs my quiet peace.)
 * The {@code configure} element and its children are then dropped by the parser
 * during the transformation and are added to a list of unknown features. Recovering
 * such comments is one of the two tasks of this handler.</p>
 * 
 * <p>The second task is recognizing links between comments and model elements. Such
 * links indicate that a comment is providing additional information for a model
 * element, and should be preserved. In the MoML file, such a link is represented by
 * the following property:</p>
 * 
 * <pre>
 *   &lt;property
 *         name="_location"
 *         class="ptolemy.kernel.util.RelativeLocation"
 *         value="[-195.0, -80.0]"&gt;
 * 
 *         &lt;property
 *               name="relativeTo"
 *               class="ptolemy.kernel.util.StringAttribute"
 *               value="Const"/&gt;
 *         &lt;property
 *               name="relativeToElementName"
 *               class="ptolemy.kernel.util.StringAttribute"
 *               value="entity"/&gt;
 *   &lt;/property&gt;
 * </pre>
 * 
 * <p>That is, the annotation's {@code _location} property is extended by two additional
 * properties that define that the location is to be interpreted relative to a given
 * model element. So far, we only support linking comments to entities, but that could
 * well be changed.</p>
 * 
 * <p>We start by extracting comments and turning them into nodes. After that's done,
 * the model is traversed in an attempt to link comments to the model elements they are
 * most likely referring to. There are two cases here:</p>
 * 
 * <ol>
 *   <li>The model contains explicit links between at least one comment and model
 *   element. In this case, we assume that the model developer added all the links he
 *   wants explicitly and only transform such links into the our way of representing
 *   them.</li>
 * 
 *   <li>The model does not contain any such explicit link. In this case, a heuristic
 *   is applied to each comment, trying to find the entity it is probably referring to.
 *   The heuristic simply looks for the nearest entity and checks if the distance is
 *   within a certain threshold value.
 * </ol>
 * 
 * <p>The heuristic can be applied in the first case as well. The explicit links will not
 * be touched, then; only unlinked comments will be considered for attachment.</p>
 * 
 * <p>The attachment heuristic is based on the distance between comments and actors.
 * Usually, the nearest actor will be chosen for attachment, up to a certain maximum
 * distance. The maximum distance was carefully chosen after an experiment.</p>
 * 
 * @author cds
 */
@SuppressWarnings("all")
public class CommentsExtractor {
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Utility class for attaching the correct rendering to comment edges.
   */
  @Inject
  @Extension
  private KRenderingFigureProvider _kRenderingFigureProvider;
  
  /**
   * If true, the attachment heuristics are disabled once explicitly attached comments are found.
   */
  private final boolean heuristicsOverride = false;
  
  /**
   * The maximum distance between a comment node and a regular node for them to be considered to be
   * attached by the comment attachment heuristic. This value is a trade-off between generating false
   * attachments and losing correct attachments.
   */
  private final double maxAttachmentDistance = 500.0;
  
  /**
   * List of comment nodes created in the process.
   */
  private final List<KNode> createdCommentNodes = CollectionLiterals.<KNode>newLinkedList();
  
  /**
   * List of explicit attachments from comment nodes to other nodes.
   */
  private final List<Pair<KNode, KNode>> explicitAttachments = CollectionLiterals.<Pair<KNode, KNode>>newLinkedList();
  
  /**
   * List of heuristically found attachments from comment nodes to other nodes.
   */
  private final List<Pair<KNode, KNode>> heuristicAttachments = CollectionLiterals.<Pair<KNode, KNode>>newLinkedList();
  
  /**
   * Finds comments and comment attachments in the tree rooted at the given node.
   * 
   * @param root the root node.
   * @param diagramSynthesis the diagram synthesis that uses this class; used to map Ptolemy model
   *                         objects to the nodes created for them.
   * @param enableAttachmentHeuristic {@code true} if the distance-based heuristic should be used to
   *                                  try and find the actors comments should be attached to. If
   *                                  this is {@code false}, only explicit attachments are extracted.
   */
  public void extractAndAttachComments(final KNode root, final AbstractDiagramSynthesis<?> diagramSynthesis, final boolean enableAttachmentHeuristic) {
    this.extractComments(root, diagramSynthesis);
    this.attachComments(enableAttachmentHeuristic);
  }
  
  /**
   * Extracts the comments attached as annotations to the given node and turns them into proper
   * nodes that are children of the given node.
   * 
   * @param root the root node.
   * @param diagramSynthesis the diagram synthesis that uses this class; used to map Ptolemy model
   *                         objects to the nodes created for them.
   */
  public void extractComments(final KNode root, final AbstractDiagramSynthesis<?> diagramSynthesis) {
    List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(root);
    for (final PropertyType annotation : _annotations) {
      String _elvis = null;
      String _class_ = annotation.getClass_();
      if (_class_ != null) {
        _elvis = _class_;
      } else {
        _elvis = "";
      }
      boolean _equals = _elvis.equals(TransformationConstants.ANNOTATION_TYPE_TEXT_ATTRIBUTE);
      if (_equals) {
        String _elvis_1 = null;
        String _annotationValue = this._annotationExtensions.getAnnotationValue(annotation, TransformationConstants.ANNOTATION_COMMENT_TEXT);
        if (_annotationValue != null) {
          _elvis_1 = _annotationValue;
        } else {
          _elvis_1 = "";
        }
        final KNode commentNode = this.addCommentNode(root, _elvis_1);
        List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(commentNode);
        List<PropertyType> _annotations_2 = this._annotationExtensions.getAnnotations(annotation);
        Iterables.<PropertyType>addAll(_annotations_1, _annotations_2);
        diagramSynthesis.<KNode>associateWith(commentNode, annotation);
      } else {
        String _elvis_2 = null;
        String _class__1 = annotation.getClass_();
        if (_class__1 != null) {
          _elvis_2 = _class__1;
        } else {
          _elvis_2 = "";
        }
        boolean _equals_1 = _elvis_2.equals(TransformationConstants.ANNOTATION_TYPE_ATTRIBUTE);
        if (_equals_1) {
          final PropertyType iconDescription = this._annotationExtensions.getAnnotation(annotation, "_iconDescription");
          boolean _notEquals = (!Objects.equal(iconDescription, null));
          if (_notEquals) {
            final String text = this.restoreCommentTextFromIconDescription(iconDescription);
            boolean _notEquals_1 = (!Objects.equal(text, null));
            if (_notEquals_1) {
              final KNode commentNode_1 = this.addCommentNode(root, text);
              List<PropertyType> _annotations_3 = this._annotationExtensions.getAnnotations(commentNode_1);
              List<PropertyType> _annotations_4 = this._annotationExtensions.getAnnotations(annotation);
              Iterables.<PropertyType>addAll(_annotations_3, _annotations_4);
              diagramSynthesis.<KNode>associateWith(commentNode_1, annotation);
            }
          }
        }
      }
    }
    EList<KNode> _children = root.getChildren();
    for (final KNode child : _children) {
      EList<KNode> _children_1 = child.getChildren();
      boolean _isEmpty = _children_1.isEmpty();
      boolean _not = (!_isEmpty);
      if (_not) {
        this.extractComments(child, diagramSynthesis);
      }
    }
  }
  
  /**
   * Tries to restore the text of a comment that was not saved as a regular comment attribute, but
   * as an SVG graphic.
   * 
   * @param iconDescription the {@code _iconDescription} attribute of the attribute that might be a
   *                        comment. This is where we will look for the SVG graphic in.
   * @return the text, if it could be restored, or {@code null} if it couldn't.
   */
  private String restoreCommentTextFromIconDescription(final PropertyType iconDescription) {
    Resource _eResource = iconDescription.eResource();
    if ((!(_eResource instanceof XMLResource))) {
      return null;
    }
    Resource _eResource_1 = iconDescription.eResource();
    final XMLResource xmlResource = ((XMLResource) _eResource_1);
    EList<ConfigureType> _configure = iconDescription.getConfigure();
    boolean _isEmpty = _configure.isEmpty();
    if (_isEmpty) {
      return null;
    }
    EList<ConfigureType> _configure_1 = iconDescription.getConfigure();
    final ConfigureType configureElement = _configure_1.get(0);
    Map<EObject, AnyType> _eObjectToExtensionMap = xmlResource.getEObjectToExtensionMap();
    final AnyType unknownFeature = _eObjectToExtensionMap.get(configureElement);
    boolean _equals = Objects.equal(unknownFeature, null);
    if (_equals) {
      return null;
    }
    FeatureMap _mixed = unknownFeature.getMixed();
    final AnyType svgElement = this.findUnknownFeature(_mixed, "svg");
    boolean _equals_1 = Objects.equal(svgElement, null);
    if (_equals_1) {
      return null;
    }
    FeatureMap _mixed_1 = svgElement.getMixed();
    final AnyType textElement = this.findUnknownFeature(_mixed_1, "text");
    boolean _equals_2 = Objects.equal(textElement, null);
    if (_equals_2) {
      return null;
    }
    FeatureMap _mixed_2 = textElement.getMixed();
    boolean _isEmpty_1 = _mixed_2.isEmpty();
    if (_isEmpty_1) {
      return null;
    }
    FeatureMap _mixed_3 = textElement.getMixed();
    FeatureMap.Entry _get = _mixed_3.get(0);
    final Object data = _get.getValue();
    if ((!(data instanceof String))) {
      return null;
    } else {
      return ((String) data);
    }
  }
  
  /**
   * Iterates through the generated comment nodes and tries to attach them to the elements they were
   * initially attached to. The attachment can have been either explicit (by the model designer) or
   * implicit, by a distance-based heuristic, if it is enabled through the given parameter.
   * 
   * @param enableHeuristic {@code true} if the distance-based attachment heuristic should be
   *                        enabled.
   */
  public void attachComments(final boolean enableHeuristic) {
    for (final KNode commentNode : this.createdCommentNodes) {
      {
        final KNode explicitAttachment = this.findExplicitAttachment(commentNode);
        boolean _notEquals = (!Objects.equal(explicitAttachment, null));
        if (_notEquals) {
          Pair<KNode, KNode> _pair = new Pair<KNode, KNode>(commentNode, explicitAttachment);
          this.explicitAttachments.add(_pair);
        } else {
          boolean _and = false;
          if (!enableHeuristic) {
            _and = false;
          } else {
            boolean _or = false;
            boolean _isEmpty = this.explicitAttachments.isEmpty();
            if (_isEmpty) {
              _or = true;
            } else {
              _or = (!this.heuristicsOverride);
            }
            _and = _or;
          }
          if (_and) {
            final KNode heuristicAttachment = this.findNearestNonCommentSibling(commentNode);
            boolean _notEquals_1 = (!Objects.equal(heuristicAttachment, null));
            if (_notEquals_1) {
              Pair<KNode, KNode> _pair_1 = new Pair<KNode, KNode>(commentNode, heuristicAttachment);
              this.heuristicAttachments.add(_pair_1);
            }
          }
        }
      }
    }
    for (final Pair<KNode, KNode> attachment : this.explicitAttachments) {
      KNode _first = attachment.getFirst();
      KNode _second = attachment.getSecond();
      this.attachCommentNode(_first, _second);
    }
    boolean _and = false;
    if (!enableHeuristic) {
      _and = false;
    } else {
      boolean _or = false;
      if ((!this.heuristicsOverride)) {
        _or = true;
      } else {
        boolean _isEmpty = this.explicitAttachments.isEmpty();
        _or = _isEmpty;
      }
      _and = _or;
    }
    if (_and) {
      for (final Pair<KNode, KNode> attachment_1 : this.heuristicAttachments) {
        KNode _first_1 = attachment_1.getFirst();
        KNode _second_1 = attachment_1.getSecond();
        this.attachCommentNode(_first_1, _second_1);
      }
    }
  }
  
  /**
   * Tries to find the node the given node is explicitly attached to, if any.
   * 
   * @param commentNode the node whose explicit attachment to find.
   * @return the node the comment was explicitly attached to, or {@code null} if none could be found.
   */
  private KNode findExplicitAttachment(final KNode commentNode) {
    final PropertyType location = this._annotationExtensions.getAnnotation(commentNode, TransformationConstants.ANNOTATION_LOCATION);
    PropertyType _annotation = null;
    if (location!=null) {
      _annotation=this._annotationExtensions.getAnnotation(location, TransformationConstants.ANNOTATION_RELATIVE_TO);
    }
    final PropertyType relativeTo = _annotation;
    PropertyType _annotation_1 = null;
    if (location!=null) {
      _annotation_1=this._annotationExtensions.getAnnotation(location, TransformationConstants.ANNOTATION_RELATIVE_TO_ELEMENT_NAME);
    }
    final PropertyType relativeToElementName = _annotation_1;
    boolean _and = false;
    boolean _and_1 = false;
    boolean _and_2 = false;
    boolean _notEquals = (!Objects.equal(relativeTo, null));
    if (!_notEquals) {
      _and_2 = false;
    } else {
      boolean _notEquals_1 = (!Objects.equal(relativeToElementName, null));
      _and_2 = _notEquals_1;
    }
    if (!_and_2) {
      _and_1 = false;
    } else {
      String _value = relativeToElementName.getValue();
      boolean _notEquals_2 = (!Objects.equal(_value, null));
      _and_1 = _notEquals_2;
    }
    if (!_and_1) {
      _and = false;
    } else {
      String _value_1 = relativeToElementName.getValue();
      boolean _equals = _value_1.equals("entity");
      _and = _equals;
    }
    if (_and) {
      KNode _parent = commentNode.getParent();
      EList<KNode> _children = _parent.getChildren();
      for (final KNode sibling : _children) {
        String _name = this._labelExtensions.getName(sibling);
        String _value_2 = relativeTo.getValue();
        boolean _equals_1 = _name.equals(_value_2);
        if (_equals_1) {
          return sibling;
        }
      }
    }
    return null;
  }
  
  /**
   * Finds the node whose position is nearest to the given comment node's position that is not a
   * comment node and not further away than the maximum attachment distance.
   * 
   * @param commentNode the node whose nearest sibling to find.
   * @return the nearest sibling or {@code null} if none could be found within the maximum distance.
   */
  private KNode findNearestNonCommentSibling(final KNode commentNode) {
    final Rectangle2D.Double commentBounds = this.getPtolemyBounds(commentNode);
    double currentNearestDistance = (this.maxAttachmentDistance + 1);
    KNode currentNearestSibling = null;
    KNode _parent = commentNode.getParent();
    EList<KNode> _children = _parent.getChildren();
    for (final KNode sibling : _children) {
      boolean _isMarkedAsComment = this._markerExtensions.isMarkedAsComment(sibling);
      boolean _not = (!_isMarkedAsComment);
      if (_not) {
        Rectangle2D.Double _ptolemyBounds = this.getPtolemyBounds(sibling);
        final double distance = this.computeSquaredDistance(commentBounds, _ptolemyBounds);
        if (((distance < currentNearestDistance) && (distance <= this.maxAttachmentDistance))) {
          currentNearestDistance = distance;
          currentNearestSibling = sibling;
        }
      }
    }
    return currentNearestSibling;
  }
  
  private final int OUT_TOP_LEFT = (Rectangle2D.OUT_TOP | Rectangle2D.OUT_LEFT);
  
  private final int OUT_BOTTOM_LEFT = (Rectangle2D.OUT_BOTTOM | Rectangle2D.OUT_LEFT);
  
  private final int OUT_TOP_RIGHT = (Rectangle2D.OUT_TOP | Rectangle2D.OUT_RIGHT);
  
  private final int OUT_BOTTOM_RIGHT = (Rectangle2D.OUT_BOTTOM | Rectangle2D.OUT_RIGHT);
  
  /**
   * Compute the squared distance between the two shapes defined by the given bounds. If the two
   * shapes intersect, a distance of zero is returned.
   * 
   * <p>The current implementation simply returns the square of the euclidean distance, which is
   * perfectly fine.</p>
   * 
   * @param bounds1 the first shape.
   * @param bounds2 the second shape.
   * @return the squared distance between the two shapes.
   */
  private double computeSquaredDistance(final Rectangle2D.Double bounds1, final Rectangle2D.Double bounds2) {
    boolean _intersects = bounds1.intersects(bounds2);
    if (_intersects) {
      return 0;
    }
    final int topLeftPos = bounds2.outcode(bounds1.x, bounds1.y);
    final int bottomRightPos = bounds2.outcode((bounds1.x + bounds1.width), (bounds1.y + bounds1.height));
    boolean _and = false;
    int _bitwiseAnd = (topLeftPos & this.OUT_TOP_LEFT);
    boolean _equals = (_bitwiseAnd == this.OUT_TOP_LEFT);
    if (!_equals) {
      _and = false;
    } else {
      int _bitwiseAnd_1 = (bottomRightPos & this.OUT_TOP_LEFT);
      boolean _equals_1 = (_bitwiseAnd_1 == this.OUT_TOP_LEFT);
      _and = _equals_1;
    }
    if (_and) {
      return this.computeSquaredDistance(
        (bounds1.x + bounds1.width), 
        (bounds1.y + bounds1.height), 
        bounds2.x, 
        bounds2.y);
    } else {
      boolean _and_1 = false;
      int _bitwiseAnd_2 = (topLeftPos & this.OUT_BOTTOM_LEFT);
      boolean _equals_2 = (_bitwiseAnd_2 == this.OUT_BOTTOM_LEFT);
      if (!_equals_2) {
        _and_1 = false;
      } else {
        int _bitwiseAnd_3 = (bottomRightPos & this.OUT_BOTTOM_LEFT);
        boolean _equals_3 = (_bitwiseAnd_3 == this.OUT_BOTTOM_LEFT);
        _and_1 = _equals_3;
      }
      if (_and_1) {
        return this.computeSquaredDistance(
          (bounds1.x + bounds1.width), 
          bounds1.y, 
          bounds2.x, 
          (bounds2.y + bounds2.height));
      } else {
        boolean _and_2 = false;
        int _bitwiseAnd_4 = (topLeftPos & this.OUT_TOP_RIGHT);
        boolean _equals_4 = (_bitwiseAnd_4 == this.OUT_TOP_RIGHT);
        if (!_equals_4) {
          _and_2 = false;
        } else {
          int _bitwiseAnd_5 = (bottomRightPos & this.OUT_TOP_RIGHT);
          boolean _equals_5 = (_bitwiseAnd_5 == this.OUT_TOP_RIGHT);
          _and_2 = _equals_5;
        }
        if (_and_2) {
          return this.computeSquaredDistance(
            bounds1.x, 
            (bounds1.y + bounds1.height), 
            (bounds2.x + bounds2.width), 
            bounds2.y);
        } else {
          boolean _and_3 = false;
          int _bitwiseAnd_6 = (topLeftPos & this.OUT_BOTTOM_RIGHT);
          boolean _equals_6 = (_bitwiseAnd_6 == this.OUT_BOTTOM_RIGHT);
          if (!_equals_6) {
            _and_3 = false;
          } else {
            int _bitwiseAnd_7 = (bottomRightPos & this.OUT_BOTTOM_RIGHT);
            boolean _equals_7 = (_bitwiseAnd_7 == this.OUT_BOTTOM_RIGHT);
            _and_3 = _equals_7;
          }
          if (_and_3) {
            return this.computeSquaredDistance(
              bounds1.x, 
              bounds1.y, 
              (bounds2.x + bounds2.width), 
              (bounds2.y + bounds2.height));
          } else {
            boolean _and_4 = false;
            int _bitwiseAnd_8 = (topLeftPos & Rectangle2D.OUT_LEFT);
            boolean _notEquals = (_bitwiseAnd_8 != 0);
            if (!_notEquals) {
              _and_4 = false;
            } else {
              int _bitwiseAnd_9 = (bottomRightPos & Rectangle2D.OUT_LEFT);
              boolean _notEquals_1 = (_bitwiseAnd_9 != 0);
              _and_4 = _notEquals_1;
            }
            if (_and_4) {
              final double distance = ((bounds1.x + bounds1.width) - bounds2.x);
              return (distance * distance);
            } else {
              boolean _and_5 = false;
              int _bitwiseAnd_10 = (topLeftPos & Rectangle2D.OUT_RIGHT);
              boolean _notEquals_2 = (_bitwiseAnd_10 != 0);
              if (!_notEquals_2) {
                _and_5 = false;
              } else {
                int _bitwiseAnd_11 = (bottomRightPos & Rectangle2D.OUT_RIGHT);
                boolean _notEquals_3 = (_bitwiseAnd_11 != 0);
                _and_5 = _notEquals_3;
              }
              if (_and_5) {
                final double distance_1 = ((bounds1.x - bounds2.x) - bounds2.width);
                return (distance_1 * distance_1);
              } else {
                boolean _and_6 = false;
                int _bitwiseAnd_12 = (topLeftPos & Rectangle2D.OUT_TOP);
                boolean _notEquals_4 = (_bitwiseAnd_12 != 0);
                if (!_notEquals_4) {
                  _and_6 = false;
                } else {
                  int _bitwiseAnd_13 = (bottomRightPos & Rectangle2D.OUT_TOP);
                  boolean _notEquals_5 = (_bitwiseAnd_13 != 0);
                  _and_6 = _notEquals_5;
                }
                if (_and_6) {
                  final double distance_2 = ((bounds1.y + bounds1.height) - bounds2.y);
                  return (distance_2 * distance_2);
                } else {
                  boolean _and_7 = false;
                  int _bitwiseAnd_14 = (topLeftPos & Rectangle2D.OUT_BOTTOM);
                  boolean _notEquals_6 = (_bitwiseAnd_14 != 0);
                  if (!_notEquals_6) {
                    _and_7 = false;
                  } else {
                    int _bitwiseAnd_15 = (bottomRightPos & Rectangle2D.OUT_BOTTOM);
                    boolean _notEquals_7 = (_bitwiseAnd_15 != 0);
                    _and_7 = _notEquals_7;
                  }
                  if (_and_7) {
                    final double distance_3 = ((bounds1.y - bounds2.y) - bounds2.height);
                    return (distance_3 * distance_3);
                  }
                }
              }
            }
          }
        }
      }
    }
    return (this.maxAttachmentDistance + 1.0);
  }
  
  /**
   * Returns the square of the distance between the two points defined by the given coordinates.
   * 
   * @param x1 x coordinate of the first point.
   * @param y1 y coordinate of the first point.
   * @param x2 x coordinate of the second point.
   * @param y2 y coordinate of the second point.
   * @return the squared distance between the two points.
   */
  private double computeSquaredDistance(final double x1, final double y1, final double x2, final double y2) {
    final double deltaX = (x2 - x1);
    final double deltaY = (y2 - y1);
    return ((deltaX * deltaX) + (deltaY * deltaY));
  }
  
  /**
   * Creates a comment node and adds it to the given parent node. Also adds it to the list of
   * created comment nodes.
   * 
   * @param parent parent of the new comment node.
   * @param text the comment text to be displayed by the comment node.
   * @return the created comment node.
   */
  private KNode addCommentNode(final KNode parent, final String text) {
    final KNode commentNode = KimlUtil.createInitializedNode();
    KLayoutData _layout = this._annotationExtensions.getLayout(commentNode);
    _layout.<String>setProperty(PtolemyProperties.COMMENT_TEXT, text);
    this._markerExtensions.markAsComment(commentNode);
    EList<KNode> _children = parent.getChildren();
    _children.add(commentNode);
    this.createdCommentNodes.add(commentNode);
    return commentNode;
  }
  
  /**
   * Attaches the given comment node to the given other node.
   * 
   * @param commentNode comment node to be attached.
   * @param attachedNode other node to attach the comment node to.
   */
  private void attachCommentNode(final KNode commentNode, final KNode attachedNode) {
    final KEdge edge = KimlUtil.createInitializedEdge();
    final KRendering edgeRendering = this._kRenderingFigureProvider.createCommentEdgeRendering(edge);
    EList<KGraphData> _data = edge.getData();
    _data.add(edgeRendering);
    edge.setSource(commentNode);
    edge.setTarget(attachedNode);
  }
  
  /**
   * Checks if the feature map contains a feature of the given name and returns that.
   * 
   * @param features the feature map to search through.
   * @param name name of the feature to look for.
   * @return the feature or {@code null} if none could be found with that name.
   */
  private AnyType findUnknownFeature(final FeatureMap features, final String name) {
    for (final FeatureMap.Entry entry : features) {
      EStructuralFeature _eStructuralFeature = entry.getEStructuralFeature();
      String _name = _eStructuralFeature.getName();
      boolean _equals = _name.equals(name);
      if (_equals) {
        Object _value = entry.getValue();
        if ((_value instanceof AnyType)) {
          Object _value_1 = entry.getValue();
          return ((AnyType) _value_1);
        }
      }
    }
    return null;
  }
  
  /**
   * Retrieves the position of the given node in the original Ptolemy diagram, if any, as well as
   * its size as given by the node's shape layout. The shape layout's size fields are expected to
   * have been correctly set during the visualization step of the diagram synthesis. This method
   * is a create method because the bounds don't change and we like Xtend to cache already
   * calculated positions for us.
   * 
   * @param node the node whose location to retrieve.
   * @return the bounds or some bounds way out if there was a problem.
   */
  private Rectangle2D.Double getPtolemyBounds(final KNode node) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(node);
    final Rectangle2D.Double _result;
    synchronized (_createCache_getPtolemyBounds) {
      if (_createCache_getPtolemyBounds.containsKey(_cacheKey)) {
        return _createCache_getPtolemyBounds.get(_cacheKey);
      }
      Rectangle2D.Double _double = new Rectangle2D.Double();
      _result = _double;
      _createCache_getPtolemyBounds.put(_cacheKey, _result);
    }
    _init_getPtolemyBounds(_result, node);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, Rectangle2D.Double> _createCache_getPtolemyBounds = CollectionLiterals.newHashMap();
  
  private void _init_getPtolemyBounds(final Rectangle2D.Double bounds, final KNode node) {
    bounds.x = 2e20;
    bounds.y = 2e20;
    final PropertyType locationAnnotation = this._annotationExtensions.getAnnotation(node, TransformationConstants.ANNOTATION_LOCATION);
    boolean _equals = Objects.equal(locationAnnotation, null);
    if (_equals) {
      return;
    }
    String _value = locationAnnotation.getValue();
    final String locationString = _value.replaceAll("[\\s\\[\\]{}]+", "");
    final String[] locationArray = locationString.split(",");
    int _size = ((List<String>)Conversions.doWrapArray(locationArray)).size();
    boolean _equals_1 = (_size == 2);
    if (_equals_1) {
      try {
        String _get = locationArray[0];
        Double _valueOf = Double.valueOf(_get);
        bounds.x = (_valueOf).doubleValue();
        String _get_1 = locationArray[1];
        Double _valueOf_1 = Double.valueOf(_get_1);
        bounds.y = (_valueOf_1).doubleValue();
        final Bounds estimatedSize = PlacementUtil.estimateSize(node);
        float _width = estimatedSize.getWidth();
        bounds.width = _width;
        float _height = estimatedSize.getHeight();
        bounds.height = _height;
      } catch (final Throwable _t) {
        if (_t instanceof NumberFormatException) {
          final NumberFormatException e = (NumberFormatException)_t;
        } else {
          throw Exceptions.sneakyThrow(_t);
        }
      }
    }
    String _xifexpression = null;
    boolean _isMarkedAsComment = this._markerExtensions.isMarkedAsComment(node);
    if (_isMarkedAsComment) {
      _xifexpression = "northwest";
    } else {
      _xifexpression = "center";
    }
    final String anchorDefault = _xifexpression;
    final PropertyType anchorAnnotation = this._annotationExtensions.getAnnotation(node, TransformationConstants.ANNOTATION_ANCHOR);
    String _elvis = null;
    String _value_1 = null;
    if (anchorAnnotation!=null) {
      _value_1=anchorAnnotation.getValue();
    }
    if (_value_1 != null) {
      _elvis = _value_1;
    } else {
      _elvis = anchorDefault;
    }
    final String anchorString = _elvis;
    boolean _matched = false;
    if (!_matched) {
      if (Objects.equal(anchorString, "north")) {
        _matched=true;
        bounds.x = (bounds.x - (bounds.width / 2));
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "south")) {
        _matched=true;
        bounds.x = (bounds.x - (bounds.width / 2));
        bounds.y = (bounds.y - bounds.height);
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "west")) {
        _matched=true;
        bounds.y = (bounds.y - (bounds.height / 2));
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "east")) {
        _matched=true;
        bounds.x = (bounds.x - bounds.width);
        bounds.y = (bounds.y - (bounds.height / 2));
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "northwest")) {
        _matched=true;
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "northeast")) {
        _matched=true;
        bounds.x = (bounds.x - bounds.width);
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "southwest")) {
        _matched=true;
        bounds.y = (bounds.y - bounds.height);
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "sountheast")) {
        _matched=true;
        bounds.x = (bounds.x - bounds.width);
        bounds.y = (bounds.y - bounds.height);
      }
    }
    if (!_matched) {
      if (Objects.equal(anchorString, "southeast")) {
        _matched=true;
        bounds.x = (bounds.x - bounds.width);
        bounds.y = (bounds.y - bounds.height);
      }
    }
    if (!_matched) {
      {
        bounds.x = (bounds.x - (bounds.width / 2));
        bounds.y = (bounds.y - (bounds.height / 2));
      }
    }
  }
}
