/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation.extensions;

import com.google.common.base.Objects;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutDataFactory;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import java.util.Arrays;
import java.util.List;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.ptolemy.moml.ClassType;
import org.ptolemy.moml.EntityType;
import org.ptolemy.moml.MomlFactory;
import org.ptolemy.moml.PropertyType;

/**
 * Utility methods regarding annotations used by the Ptolemy to KGraph transformation.
 * 
 * @author cds
 * @kieler.rating yellow 2012-07-10 KI-15 cmot, grh
 * 
 * @containsExtensions ptolemy
 */
@SuppressWarnings("all")
public class AnnotationExtensions {
  /**
   * Factory used to create MoML model instances.
   */
  private final MomlFactory momlFactory = MomlFactory.eINSTANCE;
  
  /**
   * Returns the layout data of the given element.
   * 
   * @param element the object.
   * @return the shape layout or edge layout, or {@code null} if none was found.
   */
  public KLayoutData getLayout(final KGraphElement element) {
    return element.<KLayoutData>getData(KLayoutData.class);
  }
  
  /**
   * Returns the shape layout of the given element, creating it if it does not exist yet.
   * 
   * @param node the node.
   * @return the shape layout.
   */
  public KShapeLayout getSafeLayout(final KNode node) {
    KShapeLayout layout = node.<KShapeLayout>getData(KShapeLayout.class);
    boolean _equals = Objects.equal(layout, null);
    if (_equals) {
      KShapeLayout _createKShapeLayout = KLayoutDataFactory.eINSTANCE.createKShapeLayout();
      layout = _createKShapeLayout;
      EList<KGraphData> _data = node.getData();
      _data.add(layout);
    }
    return layout;
  }
  
  /**
   * Returns the annotations of the given object.
   * This particular method looks for the property in the element's KShapeLayout.
   * 
   * @param element the object.
   * @return the annotations list or {@code null} if none was found.
   */
  protected List<PropertyType> _getAnnotations(final KGraphElement element) {
    KLayoutData _layout = this.getLayout(element);
    List<PropertyType> _property = null;
    if (_layout!=null) {
      _property=_layout.<List<PropertyType>>getProperty(PtolemyProperties.PT_PROPERTIES);
    }
    return _property;
  }
  
  /**
   * Returns the annotations of the given object.
   * This particular method looks for the properties directly attached to the element.
   * 
   * @param element the object.
   * @return the annotations list or {@code null} if none was found.
   */
  protected List<PropertyType> _getAnnotations(final PropertyType element) {
    return element.getProperty();
  }
  
  /**
   * Returns the annotations of the given entity.
   * This particular method looks for the properties directly attached to the entity.
   * 
   * @param entity the entity.
   * @return the annotations list or {@code null} if none was found.
   */
  protected List<PropertyType> _getAnnotations(final EntityType entity) {
    return entity.getProperty();
  }
  
  /**
   * Returns the annotations of the given entity.
   * This particular method looks for the properties directly attached to the entity.
   * 
   * @param entity the entity.
   * @return the annotations list or {@code null} if none was found.
   */
  protected List<PropertyType> _getAnnotations(final ClassType entity) {
    return entity.getProperty();
  }
  
  /**
   * Returns the annotation with the given key from the given KGraph element.
   * 
   * @param element the KGraph element.
   * @param key the key of the annotation to return.
   * @return the annotation or {@code null} if there is none with the given key.
   */
  public PropertyType getAnnotation(final EObject element, final String key) {
    List<PropertyType> _annotations = this.getAnnotations(element);
    final Function1<PropertyType, Boolean> _function = new Function1<PropertyType, Boolean>() {
      public Boolean apply(final PropertyType p) {
        String _name = p.getName();
        return Boolean.valueOf(_name.equals(key));
      }
    };
    return IterableExtensions.<PropertyType>findFirst(_annotations, _function);
  }
  
  /**
   * Returns the value of the string annotation with the given key, if any.
   * 
   * @param element the element to fetch the annotation from.
   * @param key the annotation's key.
   * @return the annotation's value, if it exists, or the empty string if it doesn't.
   */
  public String getAnnotationValue(final EObject element, final String key) {
    PropertyType _annotation = this.getAnnotation(element, key);
    String _value = null;
    if (_annotation!=null) {
      _value=_annotation.getValue();
    }
    return _value;
  }
  
  /**
   * Returns the boolean value of the annotation with the given key, if any.
   * 
   * @param element the element to fetch the annotation from.
   * @param key the annotation's key.
   * @return {@code true} if the annotation exists and is true, {@code false} otherwise.
   */
  public boolean getAnnotationBooleanValue(final EObject element, final String key) {
    final PropertyType annotation = this.getAnnotation(element, key);
    boolean _equals = Objects.equal(annotation, null);
    if (_equals) {
      return false;
    } else {
      String _value = annotation.getValue();
      return _value.equals("true");
    }
  }
  
  /**
   * Checks if the given KGraph element has a Ptolemy annotation of the given key. This method queries
   * the shape layout of the given element.
   * 
   * @param element the KGraph element.
   * @param key key of the annotation to look for.
   * @return {@code true} if such an annotation exists, {@code false} otherwise.
   */
  public boolean hasAnnotation(final EObject element, final String key) {
    PropertyType _annotation = this.getAnnotation(element, key);
    return (!Objects.equal(_annotation, null));
  }
  
  /**
   * Adds an annotation with the given key to the given KGraph element, if no annotation with that
   * key already exists.
   * 
   * @param element the KGraph element.
   * @param key the annotation's key.
   * @return the created property.
   */
  public PropertyType addAnnotation(final EObject element, final String key) {
    PropertyType _xifexpression = null;
    boolean _hasAnnotation = this.hasAnnotation(element, key);
    boolean _not = (!_hasAnnotation);
    if (_not) {
      final PropertyType property = this.momlFactory.createPropertyType();
      property.setName(key);
      List<PropertyType> _annotations = this.getAnnotations(element);
      _annotations.add(property);
      return property;
    } else {
      _xifexpression = this.getAnnotation(element, key);
    }
    return _xifexpression;
  }
  
  /**
   * Adds an annotation with the given key and value to the given KGraph element, if no annotation
   * with that key already exists.
   * 
   * @param element the KGraph element.
   * @param key the annotation's key.
   * @param value the annotation's value.
   * @return the created property.
   */
  public PropertyType addAnnotation(final EObject element, final String key, final String value) {
    PropertyType _xifexpression = null;
    boolean _hasAnnotation = this.hasAnnotation(element, key);
    boolean _not = (!_hasAnnotation);
    if (_not) {
      final PropertyType property = this.momlFactory.createPropertyType();
      property.setName(key);
      property.setValue(value);
      List<PropertyType> _annotations = this.getAnnotations(element);
      _annotations.add(property);
      return property;
    } else {
      _xifexpression = this.getAnnotation(element, key);
    }
    return _xifexpression;
  }
  
  /**
   * Adds an annotation with the given key, type, and value to the given KGraph element, if no
   * annotation with that key already exists.
   * 
   * @param element the KGraph element.
   * @param key the annotation's key.
   * @param type the annotation's tyape.
   * @param value the annotation's value.
   * @return the created property.
   */
  public PropertyType addAnnotation(final EObject element, final String key, final String value, final String type) {
    PropertyType _xifexpression = null;
    boolean _hasAnnotation = this.hasAnnotation(element, key);
    boolean _not = (!_hasAnnotation);
    if (_not) {
      final PropertyType property = this.momlFactory.createPropertyType();
      property.setName(key);
      property.setValue(value);
      property.setClass(type);
      List<PropertyType> _annotations = this.getAnnotations(element);
      _annotations.add(property);
      return property;
    } else {
      _xifexpression = this.getAnnotation(element, key);
    }
    return _xifexpression;
  }
  
  /**
   * Removes the annotation with the given key, if any exists, from the given KGraph element.
   * 
   * @param element the KGraph element.
   * @param key the annotation's key.
   */
  public void removeAnnotation(final EObject element, final String key) {
    final List<PropertyType> annotations = this.getAnnotations(element);
    final PropertyType annotation = this.getAnnotation(element, key);
    boolean _notEquals = (!Objects.equal(annotation, null));
    if (_notEquals) {
      annotations.remove(annotation);
    }
  }
  
  public List<PropertyType> getAnnotations(final EObject element) {
    if (element instanceof KGraphElement) {
      return _getAnnotations((KGraphElement)element);
    } else if (element instanceof ClassType) {
      return _getAnnotations((ClassType)element);
    } else if (element instanceof EntityType) {
      return _getAnnotations((EntityType)element);
    } else if (element instanceof PropertyType) {
      return _getAnnotations((PropertyType)element);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(element).toString());
    }
  }
}
