/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Kiel University
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation.extensions;

import com.google.common.base.Objects;
import com.google.common.base.Strings;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.krendering.KLeftPosition;
import de.cau.cs.kieler.core.krendering.KPosition;
import de.cau.cs.kieler.core.krendering.KRenderingFactory;
import de.cau.cs.kieler.core.krendering.KTopPosition;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutDataFactory;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.List;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.ptolemy.moml.ClassType;
import org.ptolemy.moml.EntityType;

/**
 * Utility methods used by the Ptolemy to KGraph transformation.
 * 
 * @author cds
 * @kieler.rating yellow 2012-07-10 KI-15 cmot, grh
 * 
 * @containsExtensions
 */
@SuppressWarnings("all")
public class MiscellaneousExtensions {
  /**
   * Access to annotations.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * We're using markers for some of this stuff.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Checks if the given port has an incident edge of unknown direction.
   * 
   * @param port the port to check.
   * @return {@code true} if the port has an incident edge of unknown direction,
   *         {@code false} otherwise.
   */
  public boolean hasUnknownIncidentEdge(final KPort port) {
    EList<KEdge> _edges = port.getEdges();
    for (final KEdge edge : _edges) {
      boolean _isMarkedAsUndirected = this._markerExtensions.isMarkedAsUndirected(edge);
      if (_isMarkedAsUndirected) {
        return true;
      }
    }
    return false;
  }
  
  /**
   * Reverses the given edge's direction.
   * 
   * @param edge the edge to reverse.
   */
  public void reverseEdge(final KEdge edge) {
    final KNode oldSource = edge.getSource();
    final KPort oldSourcePort = edge.getSourcePort();
    final KNode oldTarget = edge.getTarget();
    final KPort oldTargetPort = edge.getTargetPort();
    edge.setSource(oldTarget);
    edge.setSourcePort(oldTargetPort);
    edge.setTarget(oldSource);
    edge.setTargetPort(oldSourcePort);
  }
  
  /**
   * Returns the first edge in the list that is marked as having a fixed direction.
   * 
   * @param edges list of edges to check for directed edges.
   * @return the first edge of known direction that is found or {@code null} if none could be found.
   */
  public KEdge getFirstDirectedEdge(final Iterable<KEdge> edges) {
    for (final KEdge edge : edges) {
      boolean _isMarkedAsUndirected = this._markerExtensions.isMarkedAsUndirected(edge);
      boolean _not = (!_isMarkedAsUndirected);
      if (_not) {
        return edge;
      }
    }
    return null;
  }
  
  /**
   * Returns a list of all edges incident to the given node.
   * 
   * @param node the node whose incident edges to return.
   * @return a list of all incident edges.
   */
  public List<KEdge> getIncidentEdges(final KNode node) {
    final List<KEdge> incidentEdges = CollectionLiterals.<KEdge>newArrayList();
    EList<KEdge> _incomingEdges = node.getIncomingEdges();
    incidentEdges.addAll(_incomingEdges);
    EList<KEdge> _outgoingEdges = node.getOutgoingEdges();
    incidentEdges.addAll(_outgoingEdges);
    return incidentEdges;
  }
  
  /**
   * Returns the layout information of the given graph element. This method is guaranteed to not
   * return {@code null}: if the element has no layout information yet, a new layout instance is
   * created and added to the element, depending on the element's type.
   * <p>
   * chsch: renamed in order to avoid name clashes with {@link AnnotationExtensions#getLayout()}.
   * </p>
   * 
   * @param element the element whose layout information to return.
   * @return layout information of the element.
   */
  public KLayoutData getLayoutEx(final KGraphElement element) {
    if ((element instanceof KEdge)) {
      KEdgeLayout edgeLayout = ((KEdge)element).<KEdgeLayout>getData(KEdgeLayout.class);
      boolean _equals = Objects.equal(edgeLayout, null);
      if (_equals) {
        KEdgeLayout _createKEdgeLayout = KLayoutDataFactory.eINSTANCE.createKEdgeLayout();
        edgeLayout = _createKEdgeLayout;
        EList<KGraphData> _data = ((KEdge)element).getData();
        _data.add(edgeLayout);
      }
      return edgeLayout;
    } else {
      KShapeLayout shapeLayout = element.<KShapeLayout>getData(KShapeLayout.class);
      boolean _equals_1 = Objects.equal(shapeLayout, null);
      if (_equals_1) {
        KShapeLayout _createKShapeLayout = KLayoutDataFactory.eINSTANCE.createKShapeLayout();
        shapeLayout = _createKShapeLayout;
        EList<KGraphData> _data_1 = element.getData();
        _data_1.add(shapeLayout);
      }
      return shapeLayout;
    }
  }
  
  /**
   * Given a modal model state, tries to find the entity that represents its refinement, if any.
   * 
   * @param ptState the modal model state.
   * @return the entity that defines its refinement, or {@code null} if there is none or if none could
   *         be found.
   */
  public EntityType findRefinement(final EObject ptState) {
    final String refinementName = this._annotationExtensions.getAnnotationValue(ptState, TransformationConstants.ANNOTATION_REFINEMENT_NAME);
    boolean _isNullOrEmpty = Strings.isNullOrEmpty(refinementName);
    boolean _not = (!_isNullOrEmpty);
    if (_not) {
      String _xifexpression = null;
      EObject _eContainer = ptState.eContainer();
      if ((_eContainer instanceof EntityType)) {
        EObject _eContainer_1 = ptState.eContainer();
        String _class1 = ((EntityType) _eContainer_1).getClass1();
        _xifexpression = Strings.nullToEmpty(_class1);
      } else {
        String _xifexpression_1 = null;
        EObject _eContainer_2 = ptState.eContainer();
        if ((_eContainer_2 instanceof ClassType)) {
          EObject _eContainer_3 = ptState.eContainer();
          String _extends = ((ClassType) _eContainer_3).getExtends();
          _xifexpression_1 = Strings.nullToEmpty(_extends);
        }
        _xifexpression = _xifexpression_1;
      }
      final String containerClass = _xifexpression;
      boolean _or = false;
      boolean _equals = containerClass.equals(TransformationConstants.ENTITY_CLASS_MODEL_CONTROLLER);
      if (_equals) {
        _or = true;
      } else {
        boolean _equals_1 = containerClass.equals(TransformationConstants.ENTITY_CLASS_FSM_MODEL_CONTROLLER);
        _or = _equals_1;
      }
      if (_or) {
        EObject _eContainer_4 = ptState.eContainer();
        final EObject modalModel = _eContainer_4.eContainer();
        String _xifexpression_2 = null;
        if ((modalModel instanceof EntityType)) {
          String _class1_1 = ((EntityType) modalModel).getClass1();
          _xifexpression_2 = Strings.nullToEmpty(_class1_1);
        } else {
          String _xifexpression_3 = null;
          if ((modalModel instanceof ClassType)) {
            String _extends_1 = ((ClassType) modalModel).getExtends();
            _xifexpression_3 = Strings.nullToEmpty(_extends_1);
          }
          _xifexpression_2 = _xifexpression_3;
        }
        final String modalModelClass = _xifexpression_2;
        boolean _or_1 = false;
        boolean _equals_2 = modalModelClass.equals(TransformationConstants.ENTITY_CLASS_MODAL_MODEL);
        if (_equals_2) {
          _or_1 = true;
        } else {
          boolean _equals_3 = modalModelClass.equals(TransformationConstants.ENTITY_CLASS_FSM_MODAL_MODEL);
          _or_1 = _equals_3;
        }
        if (_or_1) {
          EList<EntityType> _xifexpression_4 = null;
          if ((modalModel instanceof EntityType)) {
            _xifexpression_4 = ((EntityType) modalModel).getEntity();
          } else {
            EList<EntityType> _xifexpression_5 = null;
            if ((modalModel instanceof ClassType)) {
              _xifexpression_5 = ((ClassType) modalModel).getEntity();
            }
            _xifexpression_4 = _xifexpression_5;
          }
          final EList<EntityType> childEntities = _xifexpression_4;
          boolean _notEquals = (!Objects.equal(childEntities, null));
          if (_notEquals) {
            final Function1<EntityType, Boolean> _function = new Function1<EntityType, Boolean>() {
              public Boolean apply(final EntityType child) {
                String _name = child.getName();
                return Boolean.valueOf(_name.equals(refinementName));
              }
            };
            return IterableExtensions.<EntityType>findFirst(childEntities, _function);
          }
        }
      }
    }
    return null;
  }
  
  /**
   * Creates a position with the given absolute x and y coordinates.
   * 
   * @param x the x coordinate.
   * @param y the y coordinate.
   * @return KPosition representing the given absolut coordinates.
   */
  public KPosition createKPosition(final float x, final float y) {
    final KLeftPosition xPos = KRenderingFactory.eINSTANCE.createKLeftPosition();
    xPos.setAbsolute(x);
    final KTopPosition yPos = KRenderingFactory.eINSTANCE.createKTopPosition();
    yPos.setAbsolute(y);
    final KPosition pos = KRenderingFactory.eINSTANCE.createKPosition();
    pos.setX(xPos);
    pos.setY(yPos);
    return pos;
  }
}
